/*---------------------------------------------------------------------------*
 |           PDFlib - A library for generating PDF on the fly                |
 +---------------------------------------------------------------------------+
 |          Copyright (c) 1997-2002 PDFlib GmbH. All rights reserved.        |
 *---------------------------------------------------------------------------*/

/* $Id: pdflibdl.c,v 1.7 2003/02/28 21:54:28 tm Exp $
 *
 * C wrapper code for dynamically loading the PDFlib DLL at runtime.
 *
 * Note: This module is not supported on all platforms.
 *
 */

#include <stdlib.h>

#include "pdflibdl.h"

/* remove this to disable error messages */
#define VERBOSE

/* ---------------------------------- WIN32 ----------------------------- */

#ifdef WIN32

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winbase.h>
#undef WIN32_LEAN_AND_MEAN

#define PDF_DLLNAME			"pdflib.dll"

static void *
pdf_dlopen(const char *filename)
{
    return (void *) LoadLibrary(filename);
}

static void *
pdf_dlsym(void *handle, const char *funcname)
{
    return (void *) GetProcAddress((HINSTANCE) handle, funcname);
}

static void
pdf_dlclose(void *handle)
{
    (void) FreeLibrary((HINSTANCE) handle);
}

/* ---------------------------------- MVS ----------------------------- */

#elif defined(__MVS__)

#include <dynit.h>
#include <dll.h>

#define PDF_DLLNAME			"PDFLIB"

static void *
pdf_dlopen(const char *filename)
{
    return dllload(filename);
}

static void *
pdf_dlsym(void *handle, const char *funcname)
{
    return dllqueryfn((dllhandle *) handle, funcname);
}

static void pdf_dlclose(void *handle)
{
    (void) dllfree((dllhandle *) handle);
}

/* ---------------------------------- Linux  ----------------------------- */

#elif defined(linux)

#include <dlfcn.h>

#define PDF_DLLNAME			"libpdf.so"

static void *
pdf_dlopen(const char *filename)
{
    return dlopen(filename, RTLD_LAZY);
}

static void *
pdf_dlsym(void *handle, const char *funcname)
{
    return dlsym(handle, funcname);
}

static void pdf_dlclose(void *handle)
{
    (void) dlclose(handle);
}

/* ---------------------------------- Mac OS X  ----------------------------- */

#elif defined(__ppc__) && defined(__APPLE__)

#include <mach-o/dyld.h>

#define PDF_DLLNAME			"libpdf.dylib"

static void *
pdf_dlopen(const char *filename)
{
    NSObjectFileImageReturnCode err;
    NSModule module;
    NSObjectFileImage image;

    err = NSCreateObjectFileImageFromFile(filename, &image);

    if (err != NSObjectFileImageSuccess)
	return NULL;

    module = NSLinkModule(image, filename, TRUE);
    if (module == NULL)
	return NULL;

    return module;
}

static void *
pdf_dlsym(void *handle, const char *funcname)
{
    NSSymbol symbol;
    char *name;

    (void) handle;

    name = (char*) malloc(strlen(funcname) + 2);
    sprintf(name, "_%s", funcname);
    symbol = NSLookupAndBindSymbol(name);
    free(name);
    return (void *) NSAddressOfSymbol(symbol);
}

static void pdf_dlclose(void *handle)
{
    NSUnLinkModule((NSModule) handle, FALSE);
}

/* ---------------------------------- AS/400  ----------------------------- */

#elif defined __ILEC400__

#include <string.h>

#include <miptrnam.h>
#include <qleawi.h>
#include <qusec.h>
#include "mgosifc.h"

#define PDF_DLLNAME			"PDFLIB"

static void *
pdf_dlopen(const char *filename)
{
    char libName[11], objName[11], *s;
    HMODULE handle;
    _SYSPTR pSrvPgm;
    Qle_ABP_Info_t actInfo;
    int actInfoLen;
    Qus_EC_t errCode;

    memset(libName, '\0', sizeof(libName));
    if ((s = strchr(filename, '/')) == NULL) {
	s = filename;
    } else {
	memcpy(libName, filename,  s - filename);
	if (!strcmp(libName, "*LIBL"))
	    libName[0] = '\0';
	s += 1;
    }
    strcpy(objName, s);

    /* Get system pointer to service program */
    pSrvPgm = rslvsp(WLI_SRVPGM, objName, libName, _AUTH_NONE);

    /* Activate Bound Program */
    handle = malloc(sizeof(int));
    actInfoLen = sizeof(actInfo);
    errCode.Bytes_Provided = sizeof(errCode);
    QleActBndPgm(&pSrvPgm, handle, &actInfo, &actInfoLen, &errCode);

    if (errCode.Bytes_Available > 0)
	return NULL;

    return (void *) handle;
}

static void *
pdf_dlsym(void *handle, const char *funcname)
{
    int expID;
    int expNameLen;
    void *ret;
    int expType;
    Qus_EC_t errCode;

    expID = 0;
    expNameLen = strlen(funcname);
    errCode.Bytes_Provided = sizeof(errCode);
    QleGetExp((HMODULE) handle,
	&expID, &expNameLen, funcname, &ret, &expType, &errCode);

    if (errCode.Bytes_Available > 0)
	return NULL;

    return ret;
}

static void pdf_dlclose(void *handle)
{
    free((HMODULE) handle);
}

/* ---------------------------------- unknown  ----------------------------- */

#else

#error No DLL loading code for this platform available!

#endif

/* ---------------------------------- generic  ----------------------------- */

static void
pdf_dlerror(const char *msg)
{
#ifdef VERBOSE
    fprintf(stderr, msg);
#endif
}

/* Load the PDFlib DLL and fetch the API structure */

PDFLIB_API PDFlib_api * PDFLIB_CALL
PDF_boot_dll(void)
{
    PDFlib_api *PDFlib, *(PDFLIB_CALL *get_api)(void);
    float (PDFLIB_CALL *get_value)(PDF *, const char *, float);
    float major, minor;
    void *handle;

    /* load the PDFLIB DLL... */
    handle = pdf_dlopen(PDF_DLLNAME);

    if (!handle)
    {
	pdf_dlerror("Error: couldn't load PDFlib DLL\n");
	return NULL;
    }

    /* ...and fetch function pointers */
    get_api = (PDFlib_api *(PDFLIB_CALL *)(void))
		pdf_dlsym(handle, "PDF_get_api");

    if (get_api == NULL)
    {
	pdf_dlerror(
	    "Error: couldn't find function PDF_get_api in PDFlib DLL\n");
	pdf_dlclose(handle);
	return NULL;
    }

    /* We must not yet call PDFlib->PDF_get_value() since the address
     * in the API struct may be different for an older version.
     * Therefore directly fetch the address of PDF_get_value(), and call it.
     */
    get_value = (float (PDFLIB_CALL *)(PDF *, const char *, float))
		pdf_dlsym(handle, "PDF_get_value");

    if (get_value == NULL)
    {
	pdf_dlerror(
	    "Error: couldn't find function PDF_get_value in PDFlib DLL\n");
	pdf_dlclose(handle);
	return NULL;
    }

    /*
     * Check the version numbers of the loaded DLL against that of
     * the included header file to avoid version mismatch.
     */
    major = (*get_value)(NULL, "major", 0);
    minor = (*get_value)(NULL, "minor", 0);

    if (major != PDFLIB_MAJORVERSION || minor != PDFLIB_MINORVERSION) {
	pdf_dlerror("Error: loaded wrong version of PDFlib DLL\n");
	pdf_dlclose(handle);
	return NULL;
    }

    /* Now we are safe. Fetch the API structure and boot the library. */
    PDFlib = (*get_api)();

    PDFlib->PDF_boot();

    PDFlib->handle = handle;

    return PDFlib;
}

/* unload the previously loaded PDFlib DLL */

PDFLIB_API void PDFLIB_CALL
PDF_shutdown_dll(PDFlib_api *PDFlib)
{
    if (!PDFlib || !PDFlib->handle)
	return;

    PDFlib->PDF_shutdown();

    pdf_dlclose(PDFlib->handle);
}
