/*---------------------------------------------------------------------------*
 |              PDFlib - A library for generating PDF on the fly             |
 +---------------------------------------------------------------------------+
 | Copyright (c) 1997-2003 Thomas Merz and PDFlib GmbH. All rights reserved. |
 +---------------------------------------------------------------------------+
 |                                                                           |
 |    This software is subject to the PDFlib license. It is NOT in the       |
 |    public domain. Extended versions and commercial licenses are           |
 |    available, please check http://www.pdflib.com.                         |
 |                                                                           |
 *---------------------------------------------------------------------------*/

/* $Id: pc_unicode.c,v 1.56 2003/03/14 18:13:47 tm Exp $
 *
 * PDFlib routines for converting between Unicode values and Adobe glyph names
 *
 */

#include "pc_util.h"
#include "pc_chartabs.h"


/*
 * Returns the Unicode value of a glyph name. If the name is not
 * contained in the Adobe Glyph List (AGL) 0 will be returned.
 */

pdc_ushort
pdc_adobe2unicode(const char *name)
{
    int lo = 0;
    int hi = ((sizeof tab_agl2uni) / (sizeof (pdc_glyph_tab)));

    if (name)
    {
        while (lo < hi)
        {
            int i = (lo + hi) / 2;
            int cmp = strcmp(name, tab_agl2uni[i].glyphname);

            if (cmp == 0)
                return tab_agl2uni[i].code;

            if (cmp < 0)
                hi = i;
            else
                lo = i + 1;
        }
    }

    return 0;
}

/*
 * Returns the name in the Adobe Glyph List which corresponds to
 * the supplied Unicode value. If the value doesn't have a
 * corresponding Unicode name NULL will be returned.
 */

const char *
pdc_unicode2adobe(pdc_ushort uv)
{
    int lo = 0;
    int hi = ((sizeof tab_uni2agl) / (sizeof (pdc_glyph_tab)));

    if (uv)
    {
        while (lo < hi)
        {
            int i = (lo + hi) / 2;

            if (uv == tab_uni2agl[i].code)
                return tab_uni2agl[i].glyphname;

            if (uv < tab_uni2agl[i].code)
                hi = i;
            else
                lo = i + 1;
        }
    }

    return (char *) 0;
}



/*
 * Returns true if a character name is contained in pc_standard_latin_charset.
 * Otherwise false will be returned.
 */

pdc_bool
pdc_is_std_charname(const char *name)
{
    int lo = 0;
    int hi = ((sizeof pc_standard_latin_charset) / (sizeof (char *)));

    if (name)
    {
        while (lo < hi)
        {
            int i = (lo + hi) / 2;
            int cmp = strcmp(name, pc_standard_latin_charset[i]);

            if (cmp == 0)
                return pdc_true;

            if (cmp < 0)
                hi = i;
            else
                lo = i + 1;
        }
    }

    return pdc_false;
}


/*
 *  The following source is based on Unicode's original source
 *  code ConvertUTF.c. It has been adapted to PDFlib programming
 *  conventions.
 *
 *  The original file had the following notice:
 *
 *      Copyright 2001 Unicode, Inc.
 *
 *      Limitations on Rights to Redistribute This Code
 *
 *      Author: Mark E. Davis, 1994.
 *      Rev History: Rick McGowan, fixes & updates May 2001.
 *
 *
 *  Functions for conversions between UTF32, UTF-16, and UTF-8.
 *  These funtions forming a complete set of conversions between
 *  the three formats. UTF-7 is not included here.
 *
 *  Each of these routines takes pointers to input buffers and output
 *  buffers. The input buffers are const.
 *
 *  Each routine converts the text between *sourceStart and sourceEnd,
 *  putting the result into the buffer between *targetStart and
 *  targetEnd. Note: the end pointers are *after* the last item: e.g.
 *  *(sourceEnd - 1) is the last item.
 *
 *  The return result indicates whether the conversion was successful,
 *  and if not, whether the problem was in the source or target buffers.
 *  (Only the first encountered problem is indicated.)
 *
 *  After the conversion, *sourceStart and *targetStart are both
 *  updated to point to the end of last text successfully converted in
 *  the respective buffers.
 *
 *  Input parameters:
 *      sourceStart - pointer to a pointer to the source buffer.
 *              The contents of this are modified on return so that
 *              it points at the next thing to be converted.
 *      targetStart - similarly, pointer to pointer to the target buffer.
 *      sourceEnd, targetEnd - respectively pointers to the ends of the
 *              two buffers, for overflow checking only.
 *
 *  These conversion functions take a pdc_convers_flags argument. When this
 *  flag is set to strict, both irregular sequences and isolated surrogates
 *  will cause an error.  When the flag is set to lenient, both irregular
 *  sequences and isolated surrogates are converted.
 *
 *  Whether the flag is strict or lenient, all illegal sequences will cause
 *  an error return. This includes sequences such as: <F4 90 80 80>, <C0 80>,
 *  or <A0> in UTF-8, and values above 0x10FFFF in UTF-32. Conformant code
 *  must check for illegal sequences.
 *
 *  When the flag is set to lenient, characters over 0x10FFFF are converted
 *  to the replacement character; otherwise (when the flag is set to strict)
 *  they constitute an error.
 *
 *  Output parameters:
 *      The value "sourceIllegal" is returned from some routines if the input
 *      sequence is malformed.  When "sourceIllegal" is returned, the source
 *      value will point to the illegal value that caused the problem. E.g.,
 *      in UTF-8 when a sequence is malformed, it points to the start of the
 *      malformed sequence.
 *
 *  Author: Mark E. Davis, 1994.
 *  Rev History: Rick McGowan, fixes & updates May 2001.
 *
 */

/*
 * The following 4 definitions are compiler-specific.
 * The C standard does not guarantee that wchar_t has at least
 * 16 bits, so wchar_t is no less portable than unsigned short!
 * All should be unsigned values to avoid sign extension during
 * bit mask & shift operations.
 */

typedef unsigned long   UTF32;  /* at least 32 bits */
typedef unsigned short  UTF16;  /* at least 16 bits */
typedef unsigned char   UTF8;   /* typically 8 bits */

/* Some fundamental constants */
#define UNI_SUR_HIGH_START      (UTF32)0xD800
#define UNI_SUR_HIGH_END        (UTF32)0xDBFF
#define UNI_SUR_LOW_START       (UTF32)0xDC00
#define UNI_SUR_LOW_END         (UTF32)0xDFFF
#define UNI_REPLACEMENT_CHAR    (UTF32)0x0000FFFD
#define UNI_MAX_BMP             (UTF32)0x0000FFFF
#define UNI_MAX_UTF16           (UTF32)0x0010FFFF
#define UNI_MAX_UTF32           (UTF32)0x7FFFFFFF

static const int halfShift      = 10; /* used for shifting by 10 bits */

static const UTF32 halfBase     = 0x0010000UL;
static const UTF32 halfMask     = 0x3FFUL;


/* --------------------------------------------------------------------- */

static pdc_convers_result
pdc_convertUTF32toUTF16 (
                UTF32** sourceStart, const UTF32* sourceEnd,
                UTF16** targetStart, const UTF16* targetEnd,
                const pdc_convers_flags flags) {
    pdc_convers_result result = conversionOK;
    UTF32* source = *sourceStart;
    UTF16* target = *targetStart;
    while (source < sourceEnd) {
        UTF32 ch;
        if (target >= targetEnd) {
            result = targetExhausted; break;
        }
        ch = *source++;
        if (ch <= UNI_MAX_BMP) { /* Target is a character <= 0xFFFF */
            if ((flags == strictConversion) &&
                (ch >= UNI_SUR_HIGH_START &&
                 ch <= UNI_SUR_LOW_END)) {
                --source; /* return to the illegal value itself */
                result = sourceIllegal;
                break;
            } else {
                *target++ = (UTF16) ch;     /* normal case */
            }
        } else if (ch > UNI_MAX_UTF16) {
            if (flags == strictConversion) {
                result = sourceIllegal;
            } else {
                *target++ = UNI_REPLACEMENT_CHAR;
            }
        } else {
            /* target is a character in range 0xFFFF - 0x10FFFF. */
            if (target + 1 >= targetEnd) {
                result = targetExhausted;
                break;
            }
            ch -= halfBase;
            *target++ = (UTF16) ((ch >> halfShift) + UNI_SUR_HIGH_START);
            *target++ = (UTF16) ((ch & halfMask) + UNI_SUR_LOW_START);
        }
    }
    *sourceStart = source;
    *targetStart = target;
    return result;
}

/* --------------------------------------------------------------------- */

static pdc_convers_result
pdc_convertUTF16toUTF32 (
                UTF16** sourceStart, UTF16* sourceEnd,
                UTF32** targetStart, const UTF32* targetEnd,
                const pdc_convers_flags flags) {
    pdc_convers_result result = conversionOK;
    UTF16* source = *sourceStart;
    UTF32* target = *targetStart;
    UTF32 ch, ch2;
    while (source < sourceEnd) {
        ch = *source++;
        if (ch >= UNI_SUR_HIGH_START &&
            ch <= UNI_SUR_HIGH_END &&
            source < sourceEnd) {
            ch2 = *source;
            if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END) {
                ch = ((ch - UNI_SUR_HIGH_START) << halfShift)
                      + (ch2 - UNI_SUR_LOW_START) + halfBase;
                ++source;
            } else if (flags == strictConversion) {
                /* it's an unpaired high surrogate */
                --source; /* return to the illegal value itself */
                result = sourceIllegal;
                break;
            }
        } else if ((flags == strictConversion) &&
                   (ch >= UNI_SUR_LOW_START &&
                    ch <= UNI_SUR_LOW_END)) {
            /* an unpaired low surrogate */
            --source; /* return to the illegal value itself */
            result = sourceIllegal;
            break;
        }
        if (target >= targetEnd) {
            result = targetExhausted;
            break;
        }
        *target++ = ch;
    }
    *sourceStart = source;
    *targetStart = target;
#ifdef CVTUTF_DEBUG
if (result == sourceIllegal) {
    fprintf(stderr, "pdc_convertUTF16toUTF32 illegal seq 0x%04x,%04x\n",
            ch, ch2);
    fflush(stderr);
}
#endif
    return result;
}

/* --------------------------------------------------------------------- */

/*
 * Index into the table below with the first byte of a UTF-8 sequence to
 * get the number of trailing bytes that are supposed to follow it.
 */
static const char trailingBytesForUTF8[256] = {
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
        2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

#if 0
static const char
pdc_get_trailingBytesForUTF8(int i) {
    return (trailingBytesForUTF8[i]);
}
#endif

/*
 * Magic values subtracted from a buffer value during UTF8 conversion.
 * This table contains as many values as there might be trailing bytes
 * in a UTF-8 sequence.
 */
static const UTF32 offsetsFromUTF8[6] = {
    0x00000000UL, 0x00003080UL, 0x000E2080UL,
    0x03C82080UL, 0xFA082080UL, 0x82082080UL
};

/*
 * Once the bits are split out into bytes of UTF-8, this is a mask OR-ed
 * into the first byte, depending on how many bytes follow.  There are
 * as many entries in this table as there are UTF-8 sequence types.
 * (I.e., one byte sequence, two byte... six byte sequence.)
 */
static const UTF8 firstByteMark[7] = {
    0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC
};

/* --------------------------------------------------------------------- */

/* The interface converts a whole buffer to avoid function-call overhead.
 * Constants have been gathered. Loops & conditionals have been removed as
 * much as possible for efficiency, in favor of drop-through switches.
 * (See "Note A" at the bottom of the file for equivalent code.)
 * If your compiler supports it, the "pdc_islegalUTF8" call can be turned
 * into an inline function.
 */

/* --------------------------------------------------------------------- */

static pdc_convers_result
pdc_convertUTF16toUTF8 (
                UTF16** sourceStart, const UTF16* sourceEnd,
                UTF8** targetStart, const UTF8* targetEnd,
                const pdc_convers_flags flags) {
    pdc_convers_result result = conversionOK;
    UTF16* source = *sourceStart;
    UTF8* target = *targetStart;
    while (source < sourceEnd) {
        UTF32 ch;
        unsigned short bytesToWrite = 0;
        const UTF32 byteMask = 0xBF;
        const UTF32 byteMark = 0x80;
        ch = *source++;
        /* If we have a surrogate pair, convert to UTF32 first. */
        if (ch >= UNI_SUR_HIGH_START &&
            ch <= UNI_SUR_HIGH_END &&
            source < sourceEnd) {
            UTF32 ch2 = *source;
            if (ch2 >= UNI_SUR_LOW_START &&
                ch2 <= UNI_SUR_LOW_END) {
                ch = ((ch - UNI_SUR_HIGH_START) << halfShift)
                        + (ch2 - UNI_SUR_LOW_START) + halfBase;
                ++source;
            } else if (flags == strictConversion) {
                /* it's an unpaired high surrogate */
                --source; /* return to the illegal value itself */
                result = sourceIllegal;
                break;
            }
        } else if ((flags == strictConversion) &&
                   (ch >= UNI_SUR_LOW_START &&
                    ch <= UNI_SUR_LOW_END)) {
            --source; /* return to the illegal value itself */
            result = sourceIllegal;
            break;
        }
        /* Figure out how many bytes the result will require */
        if (ch < (UTF32)0x80) {                 bytesToWrite = 1;
        } else if (ch < (UTF32)0x800) {         bytesToWrite = 2;
        } else if (ch < (UTF32)0x10000) {       bytesToWrite = 3;
        } else if (ch < (UTF32)0x200000) {      bytesToWrite = 4;
        } else {                                bytesToWrite = 2;
                                                ch = UNI_REPLACEMENT_CHAR;
        }

        target += bytesToWrite;
        if (target > targetEnd) {
            target -= bytesToWrite; result = targetExhausted; break;
        }
        switch (bytesToWrite) { /* note: everything falls through. */
            case 4: *--target = (UTF8) ((ch | byteMark) & byteMask); ch >>= 6;
            case 3: *--target = (UTF8) ((ch | byteMark) & byteMask); ch >>= 6;
            case 2: *--target = (UTF8) ((ch | byteMark) & byteMask); ch >>= 6;
            case 1: *--target = (UTF8) (ch | firstByteMark[bytesToWrite]);
        }
        target += bytesToWrite;
    }
    *sourceStart = source;
    *targetStart = target;
    return result;
}

/* --------------------------------------------------------------------- */

/*
 * Utility routine to tell whether a sequence of bytes is legal UTF-8.
 * This must be called with the length pre-determined by the first byte.
 * If not calling this from pdc_convertUTF8to*, then the length can be set by:
 *      length = trailingBytesForUTF8[*source]+1;
 * and the sequence is illegal right away if there aren't that many bytes
 * available.
 * If presented with a length > 4, this returns pdc_false.  The Unicode
 * definition of UTF-8 goes up to 4-byte sequences.
 */

static pdc_bool
pdc_islegalUTF8(UTF8 *source, int length) {
    UTF8 a;
    UTF8 *srcptr = source+length;
    switch (length) {
    default: return pdc_false;
        /* Everything else falls through when "pdc_true"... */
    case 4: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return pdc_false;
    case 3: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return pdc_false;
    case 2: if ((a = (*--srcptr)) > 0xBF) return pdc_false;
        switch (*source) {
            /* no fall-through in this inner switch */
            case 0xE0: if (a < 0xA0) return pdc_false; break;
            case 0xF0: if (a < 0x90) return pdc_false; break;
            case 0xF4: if (a > 0x8F) return pdc_false; break;
            default:  if (a < 0x80) return pdc_false;
        }
    case 1: if (*source >= 0x80 && *source < 0xC2) return pdc_false;
            if (*source > 0xF4) return pdc_false;
    }
    return pdc_true;
}

/* --------------------------------------------------------------------- */

#if 0
/*
 * Exported function to return whether a UTF-8 sequence is legal or not.
 * This is not used here; it's just exported.
 */
static pdc_bool pdc_islegalUTF8sequence(UTF8 *source, UTF8 *sourceEnd) {
    int length = trailingBytesForUTF8[*source]+1;
    if (source+length > sourceEnd) {
        return pdc_false;
    }
    return pdc_islegalUTF8(source, length);
}
#endif

/* --------------------------------------------------------------------- */

static pdc_convers_result
pdc_convertUTF8toUTF16 (
                UTF8** sourceStart, UTF8* sourceEnd,
                UTF16** targetStart, const UTF16* targetEnd,
                const pdc_convers_flags flags) {
    pdc_convers_result result = conversionOK;
    UTF8* source = *sourceStart;
    UTF16* target = *targetStart;
    while (source < sourceEnd) {
        UTF32 ch = 0L;
        unsigned short extraBytesToRead = trailingBytesForUTF8[*source];
        if (source + extraBytesToRead >= sourceEnd) {
            result = sourceExhausted;
            break;
        }
        /* Do this check whether lenient or strict */
        if (! pdc_islegalUTF8(source, extraBytesToRead+1)) {
            result = sourceIllegal;
            break;
        }
        /*
         * The cases all fall through. See "Note A" below.
         */
        switch (extraBytesToRead) {
            case 3: ch += *source++; ch <<= 6;
            case 2: ch += *source++; ch <<= 6;
            case 1: ch += *source++; ch <<= 6;
            case 0: ch += *source++;
        }
        ch -= offsetsFromUTF8[extraBytesToRead];

        if (target >= targetEnd) {
            result = targetExhausted;
            break;
        }
        if (ch <= UNI_MAX_BMP) { /* Target is a character <= 0xFFFF */
            if ((flags == strictConversion) &&
                (ch >= UNI_SUR_HIGH_START &&
                 ch <= UNI_SUR_LOW_END)) {
                --source; /* return to the illegal value itself */
                result = sourceIllegal;
                break;
            } else {
                *target++ = (UTF16) ch;     /* normal case */
            }
        } else if (ch > UNI_MAX_UTF16) {
            if (flags == strictConversion) {
                    result = sourceIllegal;
                    source -= extraBytesToRead; /* return to the start */
            } else {
                    *target++ = UNI_REPLACEMENT_CHAR;
            }
        } else {
            /* target is a character in range 0xFFFF - 0x10FFFF. */
            if (target + 1 >= targetEnd) {
                    result = targetExhausted;
                    break;
            }
            ch -= halfBase;
            *target++ = (UTF16) ((ch >> halfShift) + UNI_SUR_HIGH_START);
            *target++ = (UTF16) ((ch & halfMask) + UNI_SUR_LOW_START);
        }
    }
    *sourceStart = source;
    *targetStart = target;
    return result;
}

/* --------------------------------------------------------------------- */

static pdc_convers_result
pdc_convertUTF32toUTF8 (
                UTF32** sourceStart, const UTF32* sourceEnd,
                UTF8** targetStart, const UTF8* targetEnd,
                const pdc_convers_flags flags) {
    pdc_convers_result result = conversionOK;
    UTF32* source = *sourceStart;
    UTF8* target = *targetStart;
    while (source < sourceEnd) {
        UTF32 ch;
        unsigned short bytesToWrite = 0;
        const UTF32 byteMask = 0x000000BF;
        const UTF32 byteMark = 0x00000080;
        ch = *source++;
        /* surrogates of any stripe are not legal UTF32 characters */
        if (flags == strictConversion ) {
            if ((ch >= UNI_SUR_HIGH_START) && (ch <= UNI_SUR_LOW_END)) {
                --source; /* return to the illegal value itself */
                result = sourceIllegal;
                break;
            }
        }
        /* Figure out how many bytes the result will require */
        if (ch < (UTF32)0x80) {                 bytesToWrite = 1;
        } else if (ch < (UTF32)0x800) {         bytesToWrite = 2;
        } else if (ch < (UTF32)0x10000) {       bytesToWrite = 3;
        } else if (ch < (UTF32)0x200000) {      bytesToWrite = 4;
        } else {                                bytesToWrite = 2;
                                                ch = UNI_REPLACEMENT_CHAR;
        }

        target += bytesToWrite;
        if (target > targetEnd) {
            target -= bytesToWrite; result = targetExhausted; break;
        }
        switch (bytesToWrite) { /* note: everything falls through. */
            case 4: *--target = (UTF8) ((ch | byteMark) & byteMask); ch >>= 6;
            case 3: *--target = (UTF8) ((ch | byteMark) & byteMask); ch >>= 6;
            case 2: *--target = (UTF8) ((ch | byteMark) & byteMask); ch >>= 6;
            case 1: *--target = (UTF8) (ch | firstByteMark[bytesToWrite]);
        }
        target += bytesToWrite;
    }
    *sourceStart = source;
    *targetStart = target;
    return result;
}

/* --------------------------------------------------------------------- */

static pdc_convers_result
pdc_convertUTF8toUTF32 (
                UTF8** sourceStart, UTF8* sourceEnd,
                UTF32** targetStart, const UTF32* targetEnd,
                const pdc_convers_flags flags) {
    pdc_convers_result result = conversionOK;
    UTF8* source = *sourceStart;
    UTF32* target = *targetStart;

    (void) flags;

    while (source < sourceEnd) {
        UTF32 ch = 0;
        unsigned short extraBytesToRead = trailingBytesForUTF8[*source];
        if (source + extraBytesToRead >= sourceEnd) {
            result = sourceExhausted; break;
        }
        /* Do this check whether lenient or strict */
        if (! pdc_islegalUTF8(source, extraBytesToRead+1)) {
            result = sourceIllegal;
            break;
        }
        /*
         * The cases all fall through. See "Note A" below.
         */
        switch (extraBytesToRead) {
            case 3: ch += *source++; ch <<= 6;
            case 2: ch += *source++; ch <<= 6;
            case 1: ch += *source++; ch <<= 6;
            case 0: ch += *source++;
        }
        ch -= offsetsFromUTF8[extraBytesToRead];

        if (target >= targetEnd) {
            result = targetExhausted;
            break;
        }
        if (ch <= UNI_MAX_UTF32) {
            *target++ = ch;
        } else if (ch > UNI_MAX_UTF32) {
            *target++ = UNI_REPLACEMENT_CHAR;
        } else {
            if (target + 1 >= targetEnd) {
                result = targetExhausted;
                break;
            }
            ch -= halfBase;
            *target++ = (ch >> halfShift) + UNI_SUR_HIGH_START;
            *target++ = (ch & halfMask) + UNI_SUR_LOW_START;
        }
    }
    *sourceStart = source;
    *targetStart = target;
    return result;
}

/* ---------------------------------------------------------------------

        Note A.
        The fall-through switches in UTF-8 reading code save a
        temp variable, some decrements & conditionals.  The switches
        are equivalent to the following loop:
                {
                        int tmpBytesToRead = extraBytesToRead+1;
                        do {
                                ch += *source++;
                                --tmpBytesToRead;
                                if (tmpBytesToRead) ch <<= 6;
                        } while (tmpBytesToRead > 0);
                }
        In UTF-8 writing code, the switches on "bytesToWrite" are
        similarly unrolled loops.

   --------------------------------------------------------------------- */

/*
 * Convert a string from one UTFx format to another UTFx format
 * with various handling of BOM.
 * The function returns a pointer to an allocated buffer
 * containing the converted string.
 * The caller is responsible for freeing this buffer.
 * If return value is NULL error was occurred.
 *
 * informat
 * outformat: Key number of UTF format (8, 16, 32)
 * instring:  Pointer to input string
 * inlen:     The length of input string inclusive trailing zeroes
 * bom:       See enum pdc_bom_kind
 * *outlen:   Length of output string inclusive trailing zeroes or 0
 */
pdc_byte *
pdc_convert_string(pdc_core *pdc, int informat, int outformat,
                   const pdc_byte *instring, int inlen,
                   pdc_bom_kind bom, int *outlen)
{
    static const char *fn = "pdc_convert_string";
    int len, in = 0, out = 0;
    const pdc_byte *instr, *instre;
    pdc_byte *outstring = NULL;
    pdc_byte *outstr, *outstre;
    pdc_convers_result result = sourceIllegal;
    int issysle, isinple;

    *outlen = 0;

    if ((informat != 8 && informat != 16 && informat != 32) ||
        (outformat != 8 && outformat != 16 && outformat != 32))
        return outstring;

    /* ... better safe then sorry */
    len = inlen + 8;
    if ((informat ==  8 && (outformat == 16 || outformat == 32)) ||
        (informat == 16 && (outformat ==  8 || outformat == 32)))
        len *= 4;
    outstring = (pdc_byte *) pdc_malloc(pdc, (size_t) len, fn);

    /* is little endian system */
    issysle = !PDC_ISBIGENDIAN;
    isinple = issysle;

    /* Input string has byte order mark */
    /* (see p. 325 of The Unicode Standard Version 3.0 ) */
    if (bom > onlyConv)
    {
        switch (informat)
        {
            case 8:
            if (inlen > 3 &&
                instring[0] == PDF_BOM2 &&
                instring[1] == PDF_BOM3 &&
                instring[2] == PDF_BOM4)
                in = 3;
            break;

            case 16:
            if (inlen > 2)
            {
                if (instring[0] == PDF_BOM0 &&
                    instring[1] == PDF_BOM1)
                {
                    isinple = 0;
                    in = 2;
                }
                if (instring[0] == PDF_BOM1 &&
                    instring[1] == PDF_BOM0)
                {
                    isinple = 1;
                    in = 2;
                }
            }
            break;

            case 32:
            if (inlen > 4)
            {
                if (instring[0] == 0 &&
                    instring[1] == 0 &&
                    instring[2] == PDF_BOM0 &&
                    instring[3] == PDF_BOM1)
                {
                    isinple = 0;
                    in = 4;
                }
                if (instring[0] == PDF_BOM1 &&
                    instring[1] == PDF_BOM0 &&
                    instring[2] == 0 &&
                    instring[3] == 0)
                {
                    isinple = 1;
                    in = 4;
                }
            }
            break;
        }
    }

    /* Output string with byte order mark */
    if (bom > noBOM)
    {
        switch (outformat)
        {
            case 8:
            outstring[0] = PDF_BOM2;
            outstring[1] = PDF_BOM3;
            outstring[2] = PDF_BOM4;
            out = 3;
            break;

            case 16:
            if (bom == withBOMbe ||
                (bom == withBOMSys && !issysle) ||
                (bom == withBOMInp && !isinple))
            {
                outstring[0] = PDF_BOM0;
                outstring[1] = PDF_BOM1;
            }
            else
            {
                outstring[0] = PDF_BOM1;
                outstring[1] = PDF_BOM0;
            }
            out = 2;
            break;

            case 32:
            if (bom == withBOMbe ||
                (bom == withBOMSys && !issysle) ||
                (bom == withBOMInp && !isinple))
            {
                outstring[0] = 0;
                outstring[1] = 0;
                outstring[2] = PDF_BOM0;
                outstring[3] = PDF_BOM1;
            }
            else
            {
                outstring[0] = PDF_BOM1;
                outstring[1] = PDF_BOM0;
                outstring[2] = 0;
                outstring[3] = 0;
            }
            out = 4;
            break;
        }
    }
    instr = &instring[in];
    instre = instring + inlen;
    outstr = &outstring[out];
    outstre = outstring + len;
    result = conversionOK;

    /* Convert */
    switch (informat)
    {
        case 8:
        switch (outformat)
        {
            case 8:
            len = inlen - in;
            memcpy(outstr, instr, (size_t) len);
            outstr = outstring + len + out;
            break;

            case 16:
            result = pdc_convertUTF8toUTF16(
                         (UTF8 **) &instr, (UTF8 *) instre,
                         (UTF16 **) &outstr, (UTF16 *) outstre,
                         strictConversion);
            break;

            case 32:
            result = pdc_convertUTF8toUTF32(
                         (UTF8 **) &instr, (UTF8 *) instre,
                         (UTF32 **) &outstr, (UTF32 *) outstre,
                         strictConversion);
            break;
        }
        break;

        case 16:
        switch (outformat)
        {
            case 8:
            result = pdc_convertUTF16toUTF8(
                         (UTF16 **) &instr, (UTF16 *) instre,
                         (UTF8 **) &outstr, (UTF8 *) outstre,
                         strictConversion);
            break;

            case 16:
            len = inlen - in;
            memcpy(outstr, instr, (size_t) len);
            outstr = outstring + len + out;
            break;

            case 32:
            result = pdc_convertUTF16toUTF32(
                         (UTF16 **) &instr, (UTF16 *) instre,
                         (UTF32 **) &outstr, (UTF32 *) outstre,
                         strictConversion);
            break;
        }
        break;

        case 32:
        switch (outformat)
        {
            case 8:
            result = pdc_convertUTF32toUTF8(
                         (UTF32 **) &instr, (UTF32 *) instre,
                         (UTF8 **) &outstr, (UTF8 *) outstre,
                         strictConversion);
            break;

            case 16:
            result = pdc_convertUTF32toUTF16(
                         (UTF32 **) &instr, (UTF32 *) instre,
                         (UTF16 **) &outstr, (UTF16 *) outstre,
                         strictConversion);
            break;

            case 32:
            len = inlen - in;
            memcpy(outstr, instr, (size_t) len);
            outstr = outstring + len + out;
            break;
        }
        break;
    }

    /* Result */
    switch (result)
    {
        case conversionOK:
        *outlen = outstr - outstring;
        break;

        case targetExhausted:
        pdc_error(pdc, PDC_E_INT_UNICODEMEM, fn, 0, 0, 0);
        break;

        case sourceExhausted:
        case sourceIllegal:
        pdc_error(pdc, PDC_E_ILLARG_UTF,
                  pdc_errprintf(pdc, "%d", informat), 0, 0, 0);
        break;
    }

    if (*outlen)
    {
        outstring = (pdc_byte *)
	    pdc_realloc(pdc, outstring, (size_t) *outlen, fn);
    }
    else
    {
        pdc_free(pdc, outstring);
        outstring = NULL;
    }
    return outstring;

}

/*
 * Convert a text string in various UTF formats and probably with a BOM
 * to an 8-Bit or an UTF-16 string in machine byte order without any BOM.
 * The function returns a pointer to an allocated buffer
 * containing the converted string.
 * The caller is responsible for freeing this buffer.
 * If return value is NULL error was occurred.
 */
pdc_byte *
pdc_convert_text(pdc_core *pdc, pdc_text_format *o_inutf,
                 const pdc_byte *intext, int inlen, int *outlen)
{
    static const char *fn = "pdc_convert_text";
    pdc_text_format inutf = *o_inutf;
    pdc_ushort *usintext = (pdc_ushort *) intext;
    pdc_byte *outtext = NULL;
    int informat, i;

    if (intext)
    {
        switch (inutf)
        {
            case pdc_auto:
            case pdc_bytes:
            outtext = (pdc_byte *) pdc_malloc(pdc, (size_t) (inlen + 1), fn);
            memcpy(outtext, intext, (size_t) inlen);
            outtext[inlen] = 0;
            *outlen = inlen;
            *o_inutf = pdc_bytes;
            break;

            case pdc_auto2:
            case pdc_bytes2:
            *outlen = inlen / 2;
            for (i = 0; i < *outlen; i++)
                if (usintext[i] > 0xFF)
                    break;
            if (i == *outlen)
            {
                /* pick the low-order bytes only */
                outtext = (pdc_byte *) pdc_malloc(pdc,
                                           (size_t) (*outlen + 1), fn);
                for (i = 0; i < *outlen; i++)
                    outtext[i] = (pdc_byte) usintext[i];
                outtext[*outlen] = 0;
                *o_inutf = pdc_bytes;
                break;
            }
            inutf = pdc_utf16;

            case pdc_utf16:
#ifdef WORDS_BIGENDIAN
            if (inlen > 2 &&
                intext[0] == PDF_BOM1 &&
                intext[1] == PDF_BOM0)
                inutf = pdc_utf16le;
#else
            if (inlen > 2 &&
                intext[0] == PDF_BOM0 &&
                intext[1] == PDF_BOM1)
                inutf = pdc_utf16be;
#endif
            default:
            if (inutf == pdc_utf8)
                informat = 8;
            else
                informat = 16;
            outtext = pdc_convert_string(pdc, informat, 16, intext, inlen,
                                         noBOM, outlen);

#ifdef WORDS_BIGENDIAN
            if (inutf == pdc_utf16le && outtext)
                pdc_swap_bytes((char *) outtext, *outlen, NULL);
#else
            if (inutf == pdc_utf16be && outtext)
                pdc_swap_bytes((char *) outtext, *outlen, NULL);
#endif
            *o_inutf = pdc_utf16;
            break;
        }
    }
    return outtext;
}

/*
 * Convert a hypertext string of various UTF formats and probably with a BOM
 * to an UTF-16-BE string with BE-BOM and with 2 terminating zeros.
 * The length of the input string (inlen) must comprise terminating zeros.
 * The function returns a pointer to an allocated buffer
 * containing the converted string.
 * The caller is responsible for freeing this buffer.
 * If return value is NULL error was occurred.
 */
pdc_byte *
pdc_convert_hypertext(pdc_core *pdc, pdc_text_format inutf,
                      pdc_encodingvector *ev,
                      pdc_encodingvector *ev_pde,
                      const pdc_byte *intext, int inlen)
{
    static const char *fn = "pdc_convert_hypertext";
    pdc_ushort *usintext = (pdc_ushort *) intext;
    pdc_byte *outtext = NULL;
    pdc_byte *tmptext = NULL;
    pdc_ushort *usouttext = NULL;
    int informat, len, outlen, code = 0, i, j = -1;
    pdc_bool hasbom;
    pdc_ushort uv;

    if (intext)
    {
        switch (inutf)
        {
            case pdc_auto:
            case pdc_auto2:
            if ((inutf == pdc_auto && inlen >= 2 && pdc_is_unicode(intext)) ||
                (inutf == pdc_auto2 && inlen >= 4 &&
                 usintext[0] == (pdc_ushort) PDF_BOM0 &&
                 usintext[1] == (pdc_ushort) PDF_BOM1))
                hasbom = pdc_true;
            else
                hasbom = pdc_false;
            if (ev && !hasbom)
            {
                /* convert to PDFDocEncoding */
                inlen = (inutf == pdc_auto) ? inlen : inlen / 2;
                outtext = (pdc_byte *) pdc_malloc(pdc, (size_t) inlen + 2, fn);
                outtext[inlen] = 0;
                outtext[inlen + 1] = 0;
                for (i = 0; i < inlen; i++)
                {
                    if (inutf == pdc_auto)
                        code = (int) intext[i];
                    else
                        code = (int) usintext[i];
                    if (code)
                    {
                        if (code > 0xFF)
                        {
                            code = -1;
                            break;
                        }
                        uv = ev->codes[code];
                        if (!uv) uv = 0x0020;
                        code = pdc_get_encoding_bytecode(pdc, ev_pde, uv);
                        if (code == -1) break;
                    }
                    outtext[i] = (pdc_byte) code;
                }

                /* convert to UTF-16-BE with BOM and double null */
                if (code == -1)
                {
                    pdc_free(pdc, outtext);
                    outlen = 2 * inlen + 4;
                    outtext = (pdc_byte *) pdc_malloc(pdc, (size_t) outlen, fn);
                    outtext[++j] = PDF_BOM0;
                    outtext[++j] = PDF_BOM1;
                    for (i = 0; i < inlen; i++)
                    {
                        uv = 0;
                        if (inutf == pdc_auto)
                            code = (int) intext[i];
                        else
                            code = (int) usintext[i];
                        if (code)
                        {
                            if (code <= 0xFF)
                                uv = ev->codes[code];
                            else
                                uv = (pdc_ushort) code;
                        }
                        outtext[++j] = (pdc_byte) ((uv & 0xFF00) >> 8);
                        outtext[++j] = (pdc_byte) (uv & 0x00FF);
                    }
                    outtext[++j] = 0;
                    outtext[++j] = 0;
                }
            }
            else
            {
                if (inutf == pdc_auto)
                    inutf = pdc_bytes;
                else
                    inutf = pdc_bytes2;
            }
            break;

            default:
            break;
        }

        switch (inutf)
        {
            case pdc_auto:
            case pdc_auto2:
            break;

            case pdc_bytes:
            outtext = (pdc_byte *) pdc_malloc(pdc, (size_t) inlen + 2, fn);
            memcpy(outtext, intext, (size_t) inlen);
            outtext[inlen] = 0;
            outtext[inlen + 1] = 0;
            break;

            case pdc_bytes2:
            outlen = inlen / 2;
            for (i = 0; i < outlen; i++)
                if (usintext[i] > 0xFF)
                    break;
            if (i == outlen)
            {
                /* pick the low-order bytes only */
                outtext = (pdc_byte *) pdc_malloc(pdc, (size_t) outlen + 2, fn);
                for (i = 0; i < outlen; i++)
                    outtext[i] = (pdc_byte) usintext[i];
                outtext[i] = 0;
                outtext[i+1] = 0;
                break;
            }
            inutf = pdc_utf16;

            case pdc_utf16:
#ifdef WORDS_BIGENDIAN
            if (inlen > 2 &&
                intext[0] == PDF_BOM1 &&
                intext[1] == PDF_BOM0)
                inutf = pdc_utf16le;
#else
            if (inlen > 2 &&
                intext[0] == PDF_BOM0 &&
                intext[1] == PDF_BOM1)
                inutf = pdc_utf16be;
#endif
            default:
            if (inutf == pdc_utf8)
                informat = 8;
            else
                informat = 16;
            outtext = pdc_convert_string(pdc, informat, 16, intext, inlen,
                                         withBOMbe, &outlen);
            if (inutf == pdc_utf16le ||
                (!PDC_ISBIGENDIAN &&
                 (inutf == pdc_utf16 || inutf == pdc_utf8)))
                pdc_swap_bytes((char *) &outtext[2], outlen - 2, NULL);
            if (outtext[outlen - 1] != 0 || outtext[outlen - 2] != 0)
            {
                outlen += 2;
                outtext =
		    (pdc_byte*)pdc_realloc(pdc, outtext, (size_t) outlen, fn);
                outtext[outlen - 1] = 0;
                outtext[outlen - 2] = 0;
            }

            /* Now we have a fine Unicode hypertext. But we check
             * whether this text can be converted to PDFDocEncoding
             * text because of economical reason */
            len = (outlen - 4) / 2;
            tmptext = (pdc_byte *) pdc_malloc(pdc, (size_t) len + 1, fn);
            usouttext = (pdc_ushort *) outtext;
            for (i = 1; i <= len && ev_pde; i++)
            {
                uv = usouttext[i];
#ifndef WORDS_BIGENDIAN
                pdc_swap_bytes((char *) &uv, 2, NULL);
#endif
                code = pdc_get_encoding_bytecode(pdc, ev_pde, uv);
                if (code == -1) break;
                tmptext[i-1] = (pdc_byte) code;
            }
            if (code == -1)
            {
                pdc_free(pdc, tmptext);
            }
            else
            {
                pdc_free(pdc, outtext);
                outtext = tmptext;
                outtext[len] = 0;
            }
            break;
        }
    }
    return outtext;
}
